from typing import Optional
from frontegg.common import FronteggAuthenticator, IdentityClientMixin
from frontegg.common.frontegg_context import FronteggContext


class Frontegg:
    def __init__(self):
        self.authenticator: FronteggAuthenticator = None
        self.identity_client: IdentityClientMixin = None

    def init_app(self, client_id: str, api_key: str, options={}):
        FronteggContext.init(options)
        self.authenticator = FronteggAuthenticator(client_id, api_key)
        self.identity_client = IdentityClientMixin(self.authenticator)

    @property
    def should_refresh_vendor_token(self) -> bool:
        return self.authenticator.should_refresh_vendor_token

    @property
    def access_token(self):
        return self.authenticator.access_token

    def refresh_vendor_token(self) -> None:
        self.authenticator.refresh_vendor_token()

    @property
    def client_id(self):
        return frontegg.authenticator.client_id

    @property
    def api_key(self):
        return frontegg.authenticator.api_key

    def get_public_key(self) -> str:
        return self.identity_client.get_public_key()

    def fetch_public_key(self) -> str:
        return self.identity_client.fetch_public_key()

    def validate_identity_on_token(
            self,
            token,
            options,
            type,
    ):
        return self.identity_client.validate_identity_on_token(token, options, type)

    def decode_jwt(self, authorization_header, verify: Optional[bool] = True):
        return self.identity_client.decode_jwt(authorization_header, verify)


frontegg = Frontegg()



"""Frontegg client and proxy."""
from frontegg.helpers import exceptions as frontegg_exceptions
from frontegg.helpers.logger import logger as frontegg_logger


__all__ = ('frontegg_exceptions', 'frontegg_logger')



from typing import Optional
from frontegg.common.clients.token_resolvers.token_resolver import TokenResolver
from frontegg.common.clients.types import TokenTypes, AuthHeaderType, IEntityWithRoles, IValidateTokenOptions


class AuthorizationJWTResolver(TokenResolver[IEntityWithRoles]):
    def __init__(self):
        super().__init__([TokenTypes.TenantApiToken.value, TokenTypes.UserApiToken.value, TokenTypes.UserToken.value], AuthHeaderType.JWT.value)

    def validate_token(
            self,
            token: str,
            public_key: str,
            options: Optional[IValidateTokenOptions] = None
    ) -> IEntityWithRoles:
        entity = super().verify_token(token, public_key)

        if options is not None and (options.get('permissions') or options.get('roles')):
            self.validate_roles_and_permissions(entity, options)

        return entity

    def get_entity(self, entity: IEntityWithRoles) -> IEntityWithRoles:
        return entity



import enum
from _typeshed import Incomplete
from fastapi import Request as Request
from fastapi.security.base import SecurityBase
from frontegg.common.clients.types import AuthHeaderType as AuthHeaderType
from frontegg.helpers.exceptions import UnauthorizedException as UnauthorizedException
from frontegg.helpers.logger import logger as logger
from pydantic import BaseModel
from typing import Any, Callable, Dict, List, Optional

class TokenType(str, enum.Enum):
    UserToken: str
    UserApiToken: str
    TenantApiToken: str
    TenantAccessToken: str
    UserAccessToken: str

class User(BaseModel):
    sub: str
    roles: List[str]
    permissions: List[str]
    tenant_id: str
    token_type: TokenType
    access_token: str
    metadata: Optional[Dict[str, Any]]
    name: Optional[str]
    email: Optional[str]
    email_verified: Optional[bool]
    tenant_ids: Optional[List[str]]
    profile_picture_url: Optional[str]
    super_user: Optional[bool]
    created_by_user_id: Optional[str]
    def has_permissions(self, permissions: List[str]) -> bool: ...
    def has_roles(self, roles: List[str]) -> bool: ...
    @property
    def id(self) -> Optional[str]: ...

class FronteggHTTPAuthentication(SecurityBase):
    model: Incomplete
    scheme_name: Incomplete
    auto_error: Incomplete
    roles: Incomplete
    permissions: Incomplete
    def __init__(self, bearerFormat: Optional[str] = ..., scheme_name: Optional[str] = ..., auto_error: bool = ..., roles: List[str] = ..., permissions: List[str] = ...) -> None: ...
    def handle_authentication_failure(self) -> None: ...
    async def __call__(self, request: Request) -> Optional[User]: ...

def FronteggSecurity(permissions: List[str] = ..., auto_error: bool = ..., roles: List[str] = ...) -> Callable[[], User]: ...
def get_auth_header(req: Request) -> Optional[Dict[str, Any]]: ...



from typing import List
from frontegg.common.clients.types import IEntityWithRoles, TokenTypes, IUserAccessToken
from .access_token_service import AccessTokenService
from frontegg.common.clients.http_client import HttpClient
from frontegg.helpers.frontegg_urls import frontegg_urls
from urllib.parse import urljoin


class UserAccessTokenService(AccessTokenService[IUserAccessToken]):
    def __init__(self, client: HttpClient):
        super().__init__(TokenTypes.UserAccessToken.value)
        self.client = client

    def get_entity_from_identity(self, entity: IUserAccessToken) -> IEntityWithRoles:
        endpoint = urljoin('resources/vendor-only/users/access-tokens/v1/', entity.get('sub'))
        response = self.client.get(urljoin(frontegg_urls.identity_service['base_url'], endpoint))
        response.raise_for_status()
        data = response.json()

        return {**entity, 'roles': data.get('roles'), 'permissions': data.get('permissions')}

    def get_active_access_token_ids_from_identity(self) -> List[str]:
        response = self.client.get(
            urljoin(frontegg_urls.identity_service['base_url'], 'resources/vendor-only/users/access-tokens/v1/active'))
        response.raise_for_status()
        data = response.json()

        return data



from importlib import import_module


class PackageUtils:
    @staticmethod
    def load_package(name: str):
        try:
            return import_module(name)
        except ImportError as e:
            raise Exception(name + ' is not installed')



from functools import wraps
from frontegg.helpers.logger import logger
from time import sleep


def retry(action, total_tries=10, retry_delay=0):
    def retry_decorator(f):
        @wraps(f)
        def func_with_retries(*args, **kwargs):
            _tries = 0
            while _tries < total_tries:
                try:
                    return f(*args, **kwargs)
                except Exception as e:
                    if _tries == 0:
                        logger.info((action or f'function "{f.__name__}"') + f' failed on first try, {e}')
                    if +_tries >= total_tries - 1:
                        logger.info((action or f'function "{f.__name__}"') +
                                    f' failed on the last retry attempt ({total_tries}), {e}')
                        raise e
                    _tries += 1
                    if retry_delay > 0:
                        sleep(retry_delay)

        return func_with_retries

    return retry_decorator



from typing import Optional, List
from frontegg.common.clients.token_resolvers.token_resolver import TokenResolver
from frontegg.common.clients.token_resolvers.access_token_services.services.user_access_token_service import UserAccessTokenService
from frontegg.common.clients.token_resolvers.access_token_services.services.tenant_access_token_service import TenantAccessTokenService
from frontegg.common.clients.token_resolvers.access_token_services.cache_services.cache_user_access_token_service import CacheUserAccessTokenService
from frontegg.common.clients.token_resolvers.access_token_services.cache_services.cache_tenant_access_token_service import CacheTenantAccessTokenService
from frontegg.common.clients.types import TokenTypes, AuthHeaderType, IEntityWithRoles, IValidateTokenOptions
from frontegg.common.cache.local_cache_manager import LocalCacheManager
from frontegg.common.cache.redis_cache_manager import RedisCacheManager
from frontegg.helpers.logger import logger
from frontegg.common import FronteggAuthenticator
from frontegg.common.clients import HttpClient
from frontegg.common.frontegg_context import FronteggContext
from frontegg.helpers.exceptions import UnauthenticatedException


class AccessTokenResolver(TokenResolver[IEntityWithRoles]):
    def __init__(self, authenticator: FronteggAuthenticator):
        super().__init__([TokenTypes.TenantAccessToken.value, TokenTypes.UserAccessToken.value], AuthHeaderType.AccessToken.value)
        self.__init_access_token_services(authenticator)

    def validate_token(
            self,
            token: str,
            public_key: str,
            options: Optional[IValidateTokenOptions] = None
    ) -> IEntityWithRoles:
        entity = super().verify_token(token, public_key)
        entity_with_roles = {}

        if options is not None and ((options.get('permissions') is not None and len(options.get('permissions')) > 0) or
                                    (options.get('roles') is not None and len(options.get('roles')) > 0)):
            entity_with_roles = self.get_entity(entity)
            self.validate_roles_and_permissions(entity_with_roles, options)
        else:
            active_ids = self.__get_active_access_token_ids(entity)
            if active_ids is None or entity.get('sub') not in active_ids:
                raise UnauthenticatedException()

        return {**entity_with_roles, **entity}

    def get_entity(self, entity: IEntityWithRoles) -> List[str]:
        service = self.__get_active_access_token_service(entity.get('type'))

        return service.get_entity(entity)

    def __get_active_access_token_ids(self, entity: IEntityWithRoles) -> List[str]:
        service = self.__get_active_access_token_service(entity.get('type'))

        return service.get_active_access_token_ids()

    def __get_active_access_token_service(self, type: TokenTypes):
        resolver = None
        for _resolver in self.__access_token_services:
            if _resolver.should_handle(type):
                resolver = _resolver
                break
        if resolver is None:
            logger.error("Failed to find token resolver")
            raise UnauthenticatedException()

        return resolver

    def __init_access_token_services(self, authenticator: FronteggAuthenticator):
        http_client = HttpClient(authenticator.client_id, authenticator.api_key, '')

        cache_type = FronteggContext().options.get('access_tokens_options', {}).get('cache', {}).get('type', 'local')
        cache_options = FronteggContext().options.get('access_tokens_options', {}).get('cache', {}).get('options', {})

        if cache_type == 'redis':
            self.__access_token_services = [
                CacheUserAccessTokenService(RedisCacheManager(cache_options), RedisCacheManager(cache_options), UserAccessTokenService(http_client)),
                CacheTenantAccessTokenService(RedisCacheManager(cache_options), RedisCacheManager(cache_options), TenantAccessTokenService(http_client))
            ]
        else:
            self.__access_token_services = [
                CacheUserAccessTokenService(LocalCacheManager(), LocalCacheManager(), UserAccessTokenService(http_client)),
                CacheTenantAccessTokenService(LocalCacheManager(), LocalCacheManager(), TenantAccessTokenService(http_client))
            ]


from _typeshed import Incomplete
from frontegg.common import FronteggAuthenticator as FronteggAuthenticator, IdentityClientMixin as IdentityClientMixin
from frontegg.common.frontegg_context import FronteggContext as FronteggContext
from typing import Optional

class Frontegg:
    authenticator: Incomplete
    identity_client: Incomplete
    def __init__(self) -> None: ...
    def init_app(self, client_id: str, api_key: str, options=...): ...
    @property
    def should_refresh_vendor_token(self) -> bool: ...
    @property
    def access_token(self): ...
    def refresh_vendor_token(self) -> None: ...
    @property
    def client_id(self): ...
    @property
    def api_key(self): ...
    def get_public_key(self) -> str: ...
    def fetch_public_key(self) -> str: ...
    def validate_identity_on_token(self, token, options, type): ...
    def decode_jwt(self, authorization_header, verify: Optional[bool] = ...): ...

frontegg: Incomplete



from frontegg.common.cache.cache_manager import CacheManager
from frontegg.common.cache.cache_manager import SetOptions
from typing import List, Optional, Type, Generic, TypeVar, Tuple, Dict
from frontegg.common.package_utils import PackageUtils
import json

T = TypeVar('T')


class RedisCacheManager(Generic[T], CacheManager[T]):
    def __init__(self, options):
        redis = PackageUtils.load_package('redis')
        self.cache = redis.Redis(host=options.get('host'), port=options.get('port'), db=options.get('db'), password=options.get('password'))

    def set(self, key: str, data: Type[T], options: Optional[SetOptions] = None) -> None:
        self.cache.set(key, json.dumps(data, indent=4))
        if options and options.get('expires_in_seconds'):
            self.cache.expire(key, options.get('expires_in_seconds'))

    def get(self, key: str) -> Optional[Type[T]]:
        data = self.cache.get(key)
        if data is not None:
            return json.loads(data)

        return None

    def delete(self, keys: List[str]) -> None:
        for key in keys:
            if key in self.cache:
                del self.cache[key]



import abc
from typing import List, TypeVar, Union
from frontegg.common.clients.types import IAccessToken, IEntityWithRoles, TokenTypes, IEmptyAccessToken
from frontegg.common.cache.cache_manager import CacheManager
from frontegg.helpers.exceptions import UnauthenticatedException
from frontegg.common.clients.token_resolvers.access_token_services.base_access_token_service import \
    BaseAccessTokenService

T = TypeVar('T', bound=IAccessToken)


class CacheAccessTokenService(abc.ABC, BaseAccessTokenService[T]):
    def __init__(
            self,
            entity_cache_manager: CacheManager[Union[IEntityWithRoles, IEmptyAccessToken]],
            active_access_tokens_cache_manager: CacheManager[List[str]],
            access_token_service: BaseAccessTokenService[T],
            type: Union[TokenTypes.UserAccessToken.value, TokenTypes.TenantAccessToken.value]
    ):
        self.entity_cache_manager = entity_cache_manager
        self.active_access_tokens_cache_manager = active_access_tokens_cache_manager
        self.access_token_service = access_token_service
        self.type = type

    def get_entity(self, entity: T) -> IEntityWithRoles:
        cache_key = self.get_cache_prefix() + '_' + entity.get('sub')
        cached_data = self.entity_cache_manager.get(cache_key)

        if cached_data is not None:
            if self.__is_empty_access_token(cached_data):
                raise UnauthenticatedException()

            return cached_data

        try:
            data = self.access_token_service.get_entity(entity)
            self.entity_cache_manager.set(cache_key, data, {'expires_in_seconds': 10})

            return data
        except UnauthenticatedException:
            self.entity_cache_manager.set(cache_key, {'empty': True}, {'expires_in_seconds': 10})

    def get_active_access_token_ids(self) -> List[str]:
        cache_key = self.get_cache_prefix() + '_ids'
        cached_data = self.active_access_tokens_cache_manager.get(cache_key)

        if cached_data is not None:
            return cached_data

        try:
            data = self.access_token_service.get_active_access_token_ids()
            self.active_access_tokens_cache_manager.set(cache_key, data, {'expires_in_seconds': 10})

            return data
        except UnauthenticatedException:
            self.active_access_tokens_cache_manager.set(cache_key, [], {'expires_in_seconds': 10})

            return []

    def __is_empty_access_token(self, access_token) -> bool:
        return 'empty' in access_token and access_token['empty'] is True
    @abc.abstractmethod
    def get_cache_prefix(self, access_token) -> str:
        pass



import jwt
import os
from jwt import InvalidTokenError
from typing import Optional
from frontegg.helpers.frontegg_urls import frontegg_urls
from frontegg.helpers.logger import logger
from frontegg.helpers.retry import retry
from frontegg.common.clients.token_resolvers.authorization_header_resolver import AuthorizationJWTResolver
from frontegg.common.clients.token_resolvers.access_token_resolver import AccessTokenResolver
from frontegg.common.clients.types import AuthHeaderType, IValidateTokenOptions
from frontegg.common import FronteggAuthenticator
from frontegg.helpers.exceptions import UnauthenticatedException


jwt_decode_retry = os.environ.get('FRONTEGG_JWT_DECODE_RETRY') or '1'
jwt_decode_retry = int(jwt_decode_retry)
jwt_decode_retry_delay = os.environ.get('FRONTEGG_JWT_DECODE_RETRY_DELAY_MS') or '0'
jwt_decode_retry_delay = float(jwt_decode_retry_delay) / 1000


class IdentityClientMixin:
    __publicKey = None
    __tokenResolvers = []

    def __init__(self, authenticator: FronteggAuthenticator):
        self.__authenticator = authenticator;
        self.__tokenResolvers = [AuthorizationJWTResolver(), AccessTokenResolver(authenticator)]

    def get_public_key(self) -> str:
        if self.__publicKey:
            return self.__publicKey

        logger.info('could not find public key locally, will fetch public key')
        reties = 0
        while reties < 10:
            try:
                self.__publicKey = self.fetch_public_key()
                return self.__publicKey
            except Exception as e:
                reties = reties + 1
                logger.error(
                    'could not get public key from frontegg, retry number - ' + str(reties) + ', ' + str(e))

        logger.error('failed to get public key in all retries')

    def fetch_public_key(self) -> str:
        if self.__authenticator.should_refresh_vendor_token:
            self.__authenticator.refresh_vendor_token()

        response = self.__authenticator.vendor_session_request.get(
            frontegg_urls.identity_service['vendor_config'], timeout=3)
        response.raise_for_status()
        data = response.json()
        return data.get('publicKey')

    def validate_identity_on_token(
            self,
            token,
            options: Optional[IValidateTokenOptions] = None,
            type=AuthHeaderType.JWT.value
    ):
        if type == AuthHeaderType.JWT.value:
            try:
                token = token.replace("Bearer ", "")
            except:
                logger.error("Failed to extract token - ", token)

        public_key = None
        try:
            public_key = self.get_public_key()
        except:
            logger.error("Failed to get public key - ")
            raise UnauthenticatedException()

        resolver = None
        for _resolver in self.__tokenResolvers:
            if _resolver.should_handle(type) is True:
                resolver = _resolver
                break
        if not resolver:
            logger.error("Failed to find token resolver")
            raise UnauthenticatedException()

        entity = resolver.validate_token(token, public_key, options)
        return entity

    def decode_jwt(self, authorization_header, verify: Optional[bool] = True):
        if not authorization_header:
            raise InvalidTokenError('Authorization headers is missing')
        logger.debug('found authorization header: ' +
                     str(authorization_header))
        jwt_token = authorization_header.replace('Bearer ', '')
        public_key = self.get_public_key()
        logger.debug('got public key' + str(public_key))
        decoded = self.__get_jwt_data(jwt_token, verify, public_key)
        logger.info('jwt was decoded successfully')
        logger.debug('JWT value - ' + str(decoded))
        return decoded

    @retry(action='decode jwt', total_tries=jwt_decode_retry, retry_delay=jwt_decode_retry_delay)
    def __get_jwt_data(self, jwt_token, verify, public_key):
        if verify:
            return jwt.decode(jwt_token, public_key, algorithms=['RS256'], options={"verify_aud": False})
        return jwt.decode(jwt_token, algorithms=['RS256'], options={"verify_aud": False, "verify_signature": verify})



from fastapi.security.base import SecurityBase
from fastapi import Request, Depends, HTTPException
from typing import Optional, Any, Dict, List, Callable
import frontegg
from pydantic import BaseModel, Field
from fastapi.security.http import HTTPBearerModel
import enum
from frontegg.helpers.logger import logger
from frontegg.common.clients.types import AuthHeaderType
from frontegg.helpers.exceptions import UnauthorizedException

class TokenType(str, enum.Enum):
    UserToken = 'userToken'
    UserApiToken = 'userApiToken'
    TenantApiToken = 'tenantApiToken'
    TenantAccessToken = 'tenantAccessToken'
    UserAccessToken = 'userAccessToken'

class User(BaseModel):
    # Fields which are general for all kinds of tokens
    sub: str
    roles: List[str] = Field(default_factory=list)
    permissions: List[str] = Field(default_factory=list)
    tenant_id: str = Field(alias='tenantId')

    token_type: TokenType = Field(alias='type')
    access_token: str

    # User token fields - all fields must be optional to support API tokens
    metadata: Optional[Dict[str, Any]]
    name: Optional[str]
    email: Optional[str]
    email_verified: Optional[bool]
    tenant_ids: Optional[List[str]] = Field(alias='tenantIds', default_factory=list)
    profile_picture_url: Optional[str] = Field(alias='profilePictureUrl')
    super_user: Optional[bool] = Field(alias='superUser')

    # API Token fields - all fields must be optional to support user tokens
    created_by_user_id: Optional[str] = Field(alias='createdByUserId')

    def has_permissions(self, permissions: List[str]) -> bool:
        return bool(permissions) and all(p in self.permissions for p in permissions)

    def has_roles(self, roles: List[str]) -> bool:
        return bool(roles) and all(r in self.roles for r in roles)

    @property
    def id(self) -> Optional[str]:
        """
        When using tenant API Token, there is no user ID.
        When using user API Token, the user ID is specified in the created_by_user_id field.
        Otherwise, the user ID is specified in the sub field.
        """
        if self.token_type == TokenType.TenantApiToken:
            return None

        return self.created_by_user_id or self.sub

class FronteggHTTPAuthentication(SecurityBase):
    def __init__(self,
                 bearerFormat: Optional[str] = None,  # noqa
                 scheme_name: Optional[str] = None,
                 auto_error: bool = True,
                 roles: List[str] = [],
                 permissions: List[str] = []):
        self.model = HTTPBearerModel(bearerFormat=bearerFormat)
        self.scheme_name = scheme_name or self.__class__.__name__
        self.auto_error = auto_error
        self.roles = roles
        self.permissions = permissions

    def handle_authentication_failure(self) -> None:
        if self.auto_error:
            raise HTTPException(
                status_code=401, detail="Unauthenticated"
            )
        else:
            return None

    async def __call__(self, request: Request) -> Optional[User]:
        try:
            auth_header = get_auth_header(request)
            if auth_header is None:
                raise HTTPException(status_code=401, detail="Unauthenticated")

            decoded_user = frontegg.fastapi.frontegg.validate_identity_on_token(
                auth_header.get('token'),
                {'roles': self.roles, 'permissions': self.permissions},
                auth_header.get('type')
            )
            return User(**decoded_user, access_token=auth_header.get('token'))

        except UnauthorizedException:
            logger.info('entity does not have required role and permissions')
            raise HTTPException(status_code=403, detail='You do not have permission to perform this action.')

        except Exception as e:
            logger.error('something went wrong while validating JWT, ' + str(e))
            return self.handle_authentication_failure()

def FronteggSecurity(permissions: List[str] = [], auto_error: bool = True, roles: List[str] = []) -> Callable[[], User]:
    """
    This factory function will create an authentication dependency for FastAPI,
    and will ensure the user has the right permissions if specified.
    """

    def check_perm(user: User = Depends(FronteggHTTPAuthentication(auto_error=auto_error, roles=roles, permissions=permissions))) -> User:
        return user

    return check_perm

def get_auth_header(req: Request) -> Optional[Dict[str, Any]]:
    token = req.headers.get('Authorization')
    if token is not None:
        return {'token': token.replace('Bearer ', ''), 'type': AuthHeaderType.JWT.value}

    token = req.headers.get('x-api-key')
    if token is not None:
        return {'token': token, 'type': AuthHeaderType.AccessToken.value}

    return None



import abc
from typing import List, Optional, Type, Generic, TypeVar


class SetOptions:
    expires_in_seconds: Optional[int]


T = TypeVar('T')


class CacheManager(Generic[T], abc.ABC):
    @abc.abstractmethod
    def set(self, key: str, data: Type[T], options: Optional[SetOptions] = None) -> None:
        pass

    @abc.abstractmethod
    def get(self, key: str) -> Optional[Type[T]]:
        pass

    @abc.abstractmethod
    def delete(self, keys: List[str]) -> None:
        pass



from typing import Optional, TypedDict, List
from typing_extensions import  NotRequired
from enum import Enum
from .http_client import HttpClient


class Severity(str, Enum):
    INFO = 'Info'
    MEDIUM = 'Medium',
    HIGH = 'High',
    CRITICAL = 'Critical'
    ERROR = 'Error'


class Audit(TypedDict):
    action: NotRequired[str]
    createdAt: str
    description: NotRequired[str]
    email: NotRequired[str]
    frontegg_id: str
    ip: NotRequired[str]
    severity: Severity
    tenantId: str
    updatedAt: str
    vendorId: str


class SendAuditData(TypedDict):
    severity: Severity


class GetAuditsResponse(TypedDict):
    data: List[Audit]
    total: int


class GetAuditStatsResponse(TypedDict):
    severeThisWeek: int
    totalToday: int


def get_params(count, query_filter, filters, offset, sort_by, sort_direction):
    if not filters:
        filters = {}

    return {
        **filters,
        **{
            'filter': query_filter,
            'sortBy': sort_by,
            'sortDirection': sort_direction,
            'offset': offset,
            'count': count
        }
    }


class AuditsClient:
    def __init__(self, client: HttpClient):
        self.client = client

    def send_audit(self, audit: SendAuditData, tenant_id: str) -> dict:
        """Create new audit.

        :param tenant_id:
        :param audit: A dictionary containing the audit
        :raises requests.HTTPError: If the Frontegg API responds with an HTTP error code, this exception is raised.
        :return: The newly created audit object
        """
        response = self.client.post(
            data=audit,
            tenant_id=tenant_id
        )
        response.raise_for_status()

        return response.json()



from frontegg.common.frontegg_config import FronteggConfig
import requests
from frontegg.helpers.frontegg_urls import frontegg_urls
import arrow
from frontegg.helpers.logger import logger


class FronteggAuthenticator(FronteggConfig):
    __access_token = None
    __access_token_expiration = None
    vendor_session_request = requests.session()

    def __init__(self, client_id: str, api_key: str):
        super(FronteggAuthenticator, self).__init__(client_id, api_key)
        self.refresh_vendor_token()

    @property
    def access_token(self):
        return self.__access_token

    @property
    def should_refresh_vendor_token(self) -> bool:
        return self.__access_token is None \
               or self.__access_token_expiration is None \
               or arrow.utcnow() >= self.__access_token_expiration

    def refresh_vendor_token(self) -> None:
        body = {
            'clientId': self.client_id,
            'secret': self.api_key
        }
        logger.info('Will refresh vendor token')
        auth_url = frontegg_urls.authentication_service['authenticate_vendor']

        auth_response = self.vendor_session_request.post(auth_url, json=body, timeout=3)
        auth_response.raise_for_status()
        logger.info('Got a new vendor token from frontegg')
        response_body = auth_response.json()
        self.__access_token = response_body['token']
        self.__access_token_expiration = calcTokenExpiration(response_body['expiresIn'])
        self.vendor_session_request.headers.update({'x-access-token': self.__access_token}, timeout='3')
        logger.info('New vendor token was set successfully')


def calcTokenExpiration(expiration): return arrow.utcnow().shift(seconds=expiration * 0.8)



import typing
from functools import wraps
import frontegg.flask as __frontegg
from flask import request, abort
from frontegg.helpers.logger import logger
from flask import g
from frontegg.common.clients.types import AuthHeaderType
from frontegg.helpers.exceptions import UnauthorizedException


def with_authentication(
        permission_keys: typing.Optional[list] = None,
        role_keys: typing.Optional[list] = None
):
    def decorator(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            # Initially
            try:
                auth_header = get_auth_header(request)
                if auth_header is None:
                    abort(401)
                    return

                decoded_user = __frontegg.frontegg.validate_identity_on_token(
                    auth_header.get('token'),
                    {'roles': role_keys, 'permissions': permission_keys},
                    auth_header.get('type')
                )
                g.user = decoded_user

            except UnauthorizedException:
                logger.info('entity does not have required role and permissions')
                abort(403)

            except Exception as e:
                logger.info('something went wrong while validating JWT, ' + str(e))
                abort(401)

            logger.info('entity passed authentication middleware')

            return f(*args, **kwargs)

        return decorated_function

    return decorator


def get_auth_header(req):
    token = req.headers.get('Authorization')
    if token is not None:
        return {'token': token.replace('Bearer ', ''), 'type': AuthHeaderType.JWT.value}

    token = req.headers.get('x-api-key');
    if token is not None:
        return {'token': token, 'type': AuthHeaderType.AccessToken.value}

    return None



import os
from typing import Optional
from requests import session, Response
from ..frontegg_authenticator import FronteggAuthenticator
from urllib.parse import urljoin

timeout_in_seconds = os.environ.get('FRONTEGG_HTTP_TIMEOUT_IN_SECONDS') or '3'
timeout_in_seconds = int(timeout_in_seconds)

def merge(first: dict, second: dict):
    return {**first, **second}


def combineUrl(base_url: str, endpoint: str):
    return urljoin(base_url, endpoint)


def prepare_headers(
        tenant_id: Optional[str] = None,
        host: Optional[str] = None,
        headers: Optional[dict] = {}):
    new_headers = {}
    if tenant_id:
        new_headers['frontegg-tenant-id'] = tenant_id
    if host:
        new_headers['frontegg-vendor-host'] = host

    return merge(new_headers, headers)


class HttpClient(FronteggAuthenticator):
    def __init__(self, client_id: str, api_key: str, base_url: str):
        super(HttpClient, self).__init__(client_id=client_id, api_key=api_key)
        self.base_url = base_url
        self.client = session()

    def __prepare_auth_headers(self):
        if self.should_refresh_vendor_token:
            self.refresh_vendor_token()

        self.client.headers['x-access-token'] = self.access_token

    def get(self,
            url: str = '',
            params: Optional[dict] = None,
            tenant_id: Optional[str] = None,
            host: Optional[str] = None,
            headers: Optional[dict] = {}) -> Response:
        self.__prepare_auth_headers()
        new_headers = prepare_headers(tenant_id, host, headers)

        return self.client.get(combineUrl(self.base_url, url), params=params, headers=new_headers, timeout=timeout_in_seconds)

    def post(self,
             data,
             url: str = '',
             tenant_id: Optional[str] = None,
             host: Optional[str] = None,
             headers: Optional[dict] = {}
             ) -> Response:
        self.__prepare_auth_headers()

        new_headers = prepare_headers(tenant_id, host, headers)
        return self.client.post(combineUrl(self.base_url, url), json=data, headers=new_headers, timeout=timeout_in_seconds)

    def put(self,
            data,
            url: str = '',
            tenant_id: Optional[str] = None,
            host: Optional[str] = None,
            headers: Optional[dict] = {}
            ) -> Response:
        self.__prepare_auth_headers()

        new_headers = prepare_headers(tenant_id, host, headers)
        return self.client.put(combineUrl(self.base_url, url), json=data, headers=new_headers, timeout=timeout_in_seconds)

    def delete(self,
               url: str = '',
               tenant_id: Optional[str] = None,
               host: Optional[str] = None,
               headers: Optional[dict] = {}
               ) -> Response:
        self.__prepare_auth_headers()

        new_headers = prepare_headers(tenant_id, host, headers)
        return self.client.delete(combineUrl(self.base_url, url), headers=new_headers, timeout=timeout_in_seconds)

    def patch(self,
              data,
              url: str = '',
              tenant_id: Optional[str] = None,
              host: Optional[str] = None,
              headers: Optional[dict] = {}
              ) -> Response:
        self.__prepare_auth_headers()

        new_headers = prepare_headers(tenant_id, host, headers)
        return self.client.patch(combineUrl(self.base_url, url), json=data, headers=new_headers, timeout=timeout_in_seconds)


__all__ = 'HttpClient'



from enum import Enum
from typing import Dict, List, Optional, Union


class AuthHeaderType(Enum):
    JWT = 'JWT'
    AccessToken = 'AccessToken'


class AuthHeader:
    token: str
    type: AuthHeaderType


class IValidateTokenOptions:
    roles: Optional[List[str]] = None
    permissions: Optional[List[str]] = None


class TokenTypes(Enum):
    UserApiToken = 'userApiToken'
    TenantApiToken = 'tenantApiToken'
    UserToken = 'userToken'
    TenantAccessToken = 'tenantAccessToken'
    UserAccessToken = 'userAccessToken'


Role = str
Permission = str


class IEntity:
    id: Optional[str] = None
    sub: str
    tenantId: str
    type: TokenTypes


class IEntityWithRoles(IEntity):
    roles: List[Role]
    permissions: List[Permission]


class IEmptyAccessToken:
    empty: True


class IUser(IEntityWithRoles):
    type: TokenTypes.UserToken
    metadata: Dict[str, any]
    userId: str
    name: Optional[str] = None
    email: Optional[str] = None
    email_verified: Optional[bool] = None
    invisible: Optional[bool] = None
    tenantId: str
    superUser: Optional[bool]


class IAccessToken(IEntity):
    type: Union[TokenTypes.TenantAccessToken.value, TokenTypes.UserAccessToken.value]


class ITenantAccessToken(IAccessToken):
    type: TokenTypes.TenantAccessToken


class IUserAccessToken(IAccessToken):
    type: TokenTypes.UserAccessToken
    userId: str



from frontegg.common.clients.types import IEntityWithRoles, TokenTypes, ITenantAccessToken, IEmptyAccessToken
from .cache_access_token_service import CacheAccessTokenService
from frontegg.common.cache.cache_manager import CacheManager
from typing import Union, List
from frontegg.common.clients.token_resolvers.access_token_services.base_access_token_service import \
    BaseAccessTokenService


class CacheTenantAccessTokenService(CacheAccessTokenService[ITenantAccessToken]):

    def __init__(
        self,
        entity_cache_manager: CacheManager[Union[IEntityWithRoles, IEmptyAccessToken]],
        active_access_tokens_cache_manager: CacheManager[List[str]],
        access_token_service: BaseAccessTokenService[ITenantAccessToken],
    ):
        super().__init__(entity_cache_manager, active_access_tokens_cache_manager, access_token_service, TokenTypes.TenantAccessToken.value)

    def get_cache_prefix(self) -> str:
        return 'frontegg_sdk_v1_tenant_access_tokens'



import jwt
import abc
import os
from typing import List, Optional, Union, TypeVar, Generic
from frontegg.common.clients.types import IEntity, IEntityWithRoles, IValidateTokenOptions, TokenTypes
from frontegg.helpers.logger import logger
from frontegg.helpers.retry import retry
from frontegg.helpers.exceptions import UnauthorizedException, UnauthenticatedException

T = TypeVar('T', bound=IEntity)

jwt_decode_retry = os.environ.get('FRONTEGG_JWT_DECODE_RETRY') or '1'
jwt_decode_retry = int(jwt_decode_retry)
jwt_decode_retry_delay = os.environ.get('FRONTEGG_JWT_DECODE_RETRY_DELAY_MS') or '0'
jwt_decode_retry_delay = float(jwt_decode_retry_delay) / 1000


class TokenResolver(Generic[T], abc.ABC):
    def __init__(self, allowed_token_types: List[TokenTypes], type: str):
        self.allowed_token_types = allowed_token_types
        self.type = type

    @abc.abstractmethod
    def validate_token(self, token: str, public_key: str, options: Optional[IValidateTokenOptions] = None) -> Union[IEntity, IEntityWithRoles]:
        pass

    def verify_token(self, token: str, public_key: str) -> IEntity:
        entity = self.verify_async(token, public_key)
        self.validate_token_type(entity.get('type'))
        return entity

    def validate_token_type(self, token_type: TokenTypes):
        if token_type not in self.allowed_token_types:
            logger.info('Invalid token type')
            raise Exception()

    @staticmethod
    def validate_roles_and_permissions(entity: IEntityWithRoles, options: Optional[IValidateTokenOptions] = None):
        if options is not None:
            roles = options.get('roles')
            permissions = options.get('permissions')
            if roles is not None and len(roles) > 0:
                have_at_least_one_role = False
                for requested_role in roles:
                    if requested_role in entity.get('roles'):
                        have_at_least_one_role = True
                        break
                if not have_at_least_one_role:
                    logger.info('Insufficient role')
                    raise UnauthorizedException()

            if permissions is not None and len(permissions) > 0:
                have_at_least_one_permission = False
                for requested_permission in permissions:
                    if requested_permission in entity.get('permissions'):
                        have_at_least_one_permission = True
                        break
                if not have_at_least_one_permission:
                    logger.info('Insufficient permission')
                    raise UnauthorizedException()

    @abc.abstractmethod
    def get_entity(self, entity: IEntity) -> IEntityWithRoles:
        pass

    def should_handle(self, type: str) -> bool:
        return self.type == type

    def verify_async(self, token: str, public_key: str) -> IEntity:
        try:
            decoded = self.__get_jwt_data(token, public_key)
            return decoded
        except jwt.exceptions.DecodeError as e:
            logger.info('Failed to verify jwt - {}'.format(e))
            raise UnauthenticatedException()

    @retry(action='decode jwt', total_tries=jwt_decode_retry, retry_delay=jwt_decode_retry_delay)
    def __get_jwt_data(self, token: str, public_key: str, verify: Optional[bool] = True):
        if verify:
            return jwt.decode(token, public_key, algorithms=['RS256'], options={"verify_aud": False})
        return jwt.decode(token, algorithms=['RS256'], options={"verify_aud": False, "verify_signature": verify})



import os
from urllib.parse import urljoin
from typing import Dict


class FronteggUrls:
    __base_url = os.environ.get('FRONTEGG_API_GATEWAY_URL', 'https://api.frontegg.com/')

    def __init__(self):
        self.__authentication_base_url = os.environ.get('FRONTEGG_AUTHENTICATION_SERVICE_URL',
                                                        urljoin(self.base_url, 'auth/'))
        self.__audits_base_url = os.environ.get('FRONTEGG_AUDITS_SERVICE_URL',
                                                urljoin(self.base_url, 'audits/'))
        self.__metadata_base_url = os.environ.get('FRONTEGG_METADATA_SERVICE_URL', urljoin(self.base_url, 'metadata/'))
        self.__identity_base_url = os.environ.get('FRONTEGG_IDENTITY_SERVICE_URL', urljoin(self.base_url, 'identity/'))

    @property
    def base_url(self) -> str:
        if not self.__base_url.endswith('/'):
            self.__base_url = self.__base_url + '/'
        return self.__base_url

    @property
    def authentication_service(self) -> Dict[str, str]:
        return {
            'base_url': self.__authentication_base_url,
            'authenticate_vendor': urljoin(self.__authentication_base_url, 'vendor/')
        }

    @property
    def audits_service(self) -> Dict[str, str]:
        return {
            'base_url': self.__audits_base_url,
            'send_audits': ''
        }

    @property
    def identity_service(self) -> Dict[str, str]:
        return {
            'base_url': self.__identity_base_url,
            'vendor_config': urljoin(self.__identity_base_url, 'resources/configurations/v1/')
        }


frontegg_urls = FronteggUrls()

__all__ = 'frontegg_urls'



import abc
from typing import List, Generic, TypeVar, Union
from frontegg.common.clients.types import IAccessToken, IEntityWithRoles, TokenTypes

T = TypeVar('T', bound=IAccessToken)


class BaseAccessTokenService(Generic[T]):
    def __init__(self, type: Union[TokenTypes.UserAccessToken.value, TokenTypes.TenantAccessToken.value]):
        self.type = type

    @abc.abstractmethod
    def get_entity(self, entity: T) -> IEntityWithRoles:
        pass

    @abc.abstractmethod
    def get_active_access_token_ids(self) -> List[str]:
        pass

    def should_handle(self, type: Union[TokenTypes.UserAccessToken.value, TokenTypes.TenantAccessToken.value]) -> bool:
        return self.type == type



from frontegg.common.package_utils import PackageUtils
from frontegg.helpers.logger import logger

class FronteggContext(object):
    _instance = None
    options = {}

    def __new__(cls):
        if cls._instance is None:
            logger.debug('Creating a new frontegg context')
            cls._instance = super(FronteggContext, cls).__new__(cls)
            # Put any initialization here.
        return cls._instance

    @staticmethod
    def init(options={}):
        FronteggContext().__validate_options(options)
        FronteggContext().options = options

    def __validate_options(self, options):
        if options.get('access_tokens_options', None) is not None:
            self.__validate_access_tokens_options(options.get('access_tokens_options'))

    def __validate_access_tokens_options(self, access_tokens_options):
        if access_tokens_options.get('cache') is None:
            raise Exception("'cache' is missing from access tokens options")

        if access_tokens_options.get('cache').get('type') == 'redis':
            self.__validate_redis_options(access_tokens_options.get('cache').get('options', {}))

    def __validate_redis_options(self, redis_options):
        PackageUtils.load_package('redis')

        required_properties = ['host', 'password', 'port', 'db']
        for required_property in required_properties:
            if redis_options.get(required_property, None) is None:
                raise Exception(required_property + ' is missing from redis cache options')



from .frontegg_authenticator import FronteggAuthenticator
from .identity_mixin import IdentityClientMixin
from frontegg.common.frontegg_config import FronteggConfig

__all__ = ('FronteggAuthenticator', 'IdentityClientMixin', 'FronteggConfig')



from frontegg.common.clients.audits_client import AuditsClient, Severity, Audit
from frontegg.common.clients.http_client import HttpClient

__all__ = ('HttpClient', 'AuditsClient', 'Severity', 'Audit')



import logging
import os
logger = logging.getLogger('Frontegg')
format = logging.Formatter('[Frontegg:%(levelname)s]: %(message)s')
sh = logging.StreamHandler()
sh.setFormatter(format)
logger.addHandler(sh)

debug = os.environ.get('FRONTEGG_DEBUG')

if debug:
    logger.setLevel(logging.DEBUG)



import abc
import json
from typing import List, TypeVar, Union
from frontegg.common.clients.types import IAccessToken, IEntityWithRoles, TokenTypes
from frontegg.helpers.exceptions import UnauthenticatedException
from frontegg.helpers.logger import logger
from frontegg.common.clients.token_resolvers.access_token_services.base_access_token_service import \
    BaseAccessTokenService

T = TypeVar('T', bound=IAccessToken)


class AccessTokenService(abc.ABC, BaseAccessTokenService[T]):
    def __init__(self, type: Union[TokenTypes.UserAccessToken.value, TokenTypes.TenantAccessToken.value]):
        self.type = type

    def get_entity(self, entity: T) -> IEntityWithRoles:
        try:
            data = self.get_entity_from_identity(entity)
            return data
        except Exception as e:
            logger.exception('Failed to get entity from identity')
            if self.__is_api_tokens_disabled(e):
                raise UnauthenticatedException()

            raise e

    def get_active_access_token_ids(self) -> List[str]:
        try:
            return self.get_active_access_token_ids_from_identity()
        except Exception as e:
            logger.exception('Failed to get active access token ids')
            if self.__is_api_tokens_disabled(e):
                raise UnauthenticatedException()

            raise e

    @abc.abstractmethod
    def get_entity_from_identity(self, entity: T) -> IEntityWithRoles:
        pass

    @abc.abstractmethod
    def get_active_access_token_ids_from_identity(self) -> List[str]:
        pass

    def __is_api_tokens_disabled(self, error):
        try:
            return json.loads(error.response.text).get('errors')[0] == 'Api tokens are disabled' and error.response.status_code == 403
        except:
            return True





from frontegg.common.cache.cache_manager import CacheManager
from frontegg.common.cache.cache_manager import SetOptions
from typing import List, Optional, Type, Generic, TypeVar, Tuple, Dict
from time import time

T = TypeVar('T')


class LocalCacheManager(Generic[T], CacheManager[T]):
    def __init__(self):
        self.cache: Dict[str, Tuple[Type[T], float]] = {}

    def set(self, key: str, data: Type[T], options: Optional[SetOptions] = None) -> None:
        if options and options.get('expires_in_seconds'):
            self.cache[key] = (data, time() + options.get('expires_in_seconds'))
        else:
            self.cache[key] = data

    def get(self, key: str) -> Optional[Type[T]]:
        if key in self.cache:
            if isinstance(self.cache[key], tuple):
                if time() > self.cache[key][1]:
                    del self.cache[key]
                    return None
                else:
                    return self.cache[key][0]
            else:
                return self.cache[key]
        else:
            return None

    def delete(self, keys: List[str]) -> None:
        for key in keys:
            if key in self.cache:
                del self.cache[key]



class FronteggConfig:
    client_id: str = None
    api_key: str = None

    def __init__(self, client_id: str, api_key):
        if client_id is None:
            raise Exception('client_id is required')
        if api_key is None:
            raise Exception('api_key is required')

        self.client_id = client_id
        self.api_key = api_key
